//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gertespezifischer Funktionen und Mechanismen            //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2010 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt16           prgLdrMode __attribute__ ((section (".noinit")));

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define FreezeWhileRegBitClr(reg, bit, timout) FreezeWhileRegBitClr_D(&(reg), bit, timout)
__attribute__((always_inline)) static inline uint8_t
FreezeWhileRegBitClr_D(volatile uint8_t* sfr, uint8_t bit, uint8_t timout)
{
    uint8_t            mask;
    volatile uint16_t  bailOut;

    mask    = 1 << bit;
    bailOut = (F_CPU/(1+CKDIV8*7)*timout)/(8*1000); // 8 = # Zyklen pro Iteration
    if( bailOut == 0 ) bailOut = 1;
    __asm__ __volatile__ ("L1_%=:  ld    __tmp_reg__, x   \n\t"
                          "        and   __tmp_reg__, %2  \n\t"
                          "        brne  L2_%=            \n\t"
                          "        sbiw  %0,1             \n\t"
                          "        brne  L1_%=            \n\t"
                          "L2_%=:                             "
                          : "+w" (bailOut)
                          : "x" (sfr),
                            "r" (mask)
                          );
    return bailOut > 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Vorlauf  /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//                           _                                                          //
//                          (_)                                                         //
//                           |                                                          //
//                           |                                                          //
//                           |                                                          //
//                          \|/                                                         //
//                    /\  ja                                              //
//                   | Power-On-Reset |------------------.                              //
//                    \______________/                   |                              //
//                           | nein                      |                              //
//                           |                           |                              //
//                           |                           |                              //
//                          \|/                         \|/                             //
//                    /\ nein       /\ ja                   //
//                   | External-Reset |--------->| Watchdog-Reset |-----.               //
//                    \______________/            \______________/       |              //
//                           | ja                        | nein          |              //
//                           |                           |               |              //
//                           |                           |               |              //
//                          \|/                         \|/              |              //
//                +----------------------+   +-----------------------+   |              //
//                | mode = PGMLDR_MARKER |   | mode = !PGMLDR_MARKER |   |              //
//                +----------------------+   +-----------------------+   |              //
//                           |                           |               |              //
//                           |                           |               |              //
//                           |<-----------------------------------------              //
//                           |                                                          //
//                           |                                                          //
//                          \|/                                                         //
//                /\                                               //
//               | mode != PGMLDR_MARKER | ja    +-------------------+                  //
//               |          und          |------>| Start Applikation |                  //
//               |   Read(0) != 0xFFFF   |       +-------------------+                  //
//                \_____________________/                                               //
//                           | nein                                                     //
//                           |                                                          //
//                           |                                                          //
//                           |                                                          //
//                                                                                 //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked)) __attribute__((section(".init3"))) void
PreStart(void)
{
    UInt08 mcusr = MCUSR;
    MCUSR = 0;
    UInt16 mode  = prgLdrMode;
    wdt_enable(WDTO_2S);
    if( mcusr & BV(PORF) )
    {   mode = !CFG_PGMLDR_MARKER;
    }
    else
    {   if( mcusr & BV(EXTRF) )
        {   mode = CFG_PGMLDR_MARKER;
        }
        else
        {   if( !(mcusr & BV(WDRF)) )
            {   mode = !CFG_PGMLDR_MARKER;
            }
        }
    }
    if( mode != CFG_PGMLDR_MARKER && pgm_read_word_near(0) != 0xFFFF ) GOTO(0);  // Appl.
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(__vector_default) __attribute__((naked));
ISR(__vector_default)
{
    // Programmlader Einstiegspunkt
    DisableInterrupts();
    BOARD_Reset();
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Setup(void)
{
    // Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(A, _P_,    // PA.0: A0_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.1: A1_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.2: A2_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.3: A3_GPIO    : Allgemeiner Anschluss
                _T_,    // PA.4: A4_PDN	    : Allgemeiner Pull-Down-Anschluss
                _T_,    // PA.5: A5_PUP     : Allgemeiner Pull-Up-Anschluss
                _T_,    // PA.6: A6_AIN     : Analogeingang 0 .. 5V
                _L_);   // PA.7: A7_SYS_LED : System-Led: L = aus
    ConfPort(B, _P_,    // PB.0: B0_GPIO    : Allgemeiner Anschluss
                _T_,    // PB.1: B1_TA1     : Allgemeine Drucktaste 1, L = gedrckt
                _T_,    // PB.2: B2_TA2     : Allgemeine Drucktaste 2  L = gedrckt
                _L_,    // PB.3: B3_PIEZO   : Piezosignalgeber, Frequenz ber Timer 0 PWM
                _P_,    // PB.4: B4_GPIO    : Allgemeiner Anschluss
                _P_,    // PB.5: B5_MOSI    : SPI/ISP
                _P_,    // PB.6: B6_MISO    : SPI/ISP
                _P_);   // PB.7: B7_SCK     : SPI/ISP
    ConfPort(C, _P_,    // PC.0: C0_SCL     : TWI (IC)
                _P_,    // PC.1: C1_SDA     : TWI (IC)
                _P_,    // PC.2: C2_TCK     : JTAG
                _P_,    // PC.3: C3_TMS     : JTAG
                _P_,    // PC.4: C4_TDO     : JTAG
                _P_,    // PC.5: C5_TDI     : JTAG
                _L_,    // PC.6: C6_BUS_ACT : Cy4NET-Aktivitt, L = keine
                _P_);   // PC.7: C7_GPIO    : Allgemeiner Anschluss
    ConfPort(D, _P_,    // PD.0: RXD        : Bus-RXD
                _P_,    // PD.1: TXD        : Bus-TXD
                _P_,    // PD.2: D2_GPIO    : Allgemeiner Anschluss
                _P_,    // PD.3: D3_GPIO    : Allgemeiner Anschluss
                _P_,    // PD.4: D4_GPIO    : Allgemeiner Anschluss
                _L_,    // PD.5: D5_GP_LED3 : Allgemeine LED 3, H = Ein
                _L_,    // PD.6: D6_GP_LED2 : Allgemeine LED 2, H = Ein
                _L_);   // PD.7: D7_GP_LED1 : Allgemeine LED 1, H = Ein (PWM Timer 2)
	// Systemzustandsleuchte konfigurieren
#if CFG_LED_LOW_ACTIVE == true
    ConfPin(CFG_SYSTEM_LED, _L_);
#else
    ConfPin(CFG_SYSTEM_LED, _H_);
#endif
    // 16-Bit Timer 1 ohne IRQ zhlen lassen (F_CPU1024)
    TCCR1B = BV(CS12) | BV(CS10);
    // UART: 19200 Baud, 8N1, async
    UCSRA = 0;
    UCSRB = BV(RXEN) | BV(TXEN);
    UBRRH = (UInt08)((F_CPU/16L/19200-1)>>8);
    UBRRL = (UInt08)(F_CPU/16L/19200-1);
}

//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) void
BOARD_Reset(void)
{
    prgLdrMode = CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while( true );
}

//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) void
BOARD_EnterAppl(void)
{
    prgLdrMode = !CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while( true );
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_TransmitByte(UInt08 byte)
{
    loop_until_bit_is_set(UCSRA, UDRE);
    UDR = byte;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_WaitForByte(UInt08* byte, UInt08 timeout)
{
    if( !FreezeWhileRegBitClr(UCSRA, RXC, timeout) ) return false;
    *byte = UDR;
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////

